// shadowtex.cpp
// by Glenn G. Chappell
// December 2003
//
// For CS 381
// Demo of shadows using texture

#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff, includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 512; // Starting window width & height, in pixels
                              // Also used for texture size; must be power of 2.
const double neardist = 1.0;  // Near & far clipping distances
const double fardist = 10.0;

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// For light
GLfloat light0_maincolor[] = { 1., 1., 1., 1. };
GLfloat light0_ambcolor[] = { 0., 0., 0., 0. };
GLfloat light0_position[] = { 4., 0., -1., 1. };
   // Above are parameters for GL_LIGHT0
GLfloat shreceiverpos[] = { -4., 0., -7. };

// For texture
const int img_width = startwinsize;
const int img_height = startwinsize;
GLubyte the_image[img_height][img_width][3];
   // Temporary image storage for texture
   // 3rd subscript 0 = R, 1 = G, 2 = B
const int numtextures = 2;    // How many textures there are
GLuint texnames[numtextures]; // Names for our textures

// For user-controlled stuff
int whichtexture = 0;         // Which texture is currently being used
                              // Legal values in [0..numtextures-1]
bool textureon = true;        // True if texture enabled
bool rotate = false;          // True if shadow receiver is rotating
double angle = 0.;            // Current angle of shadow receiver
const double anglestep = 1.;  // Rotation speed (degrees/frame)
bool drawsquare = false;      // True if textured square is drawn


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// setmaterial_main
// Sets up "main" material
void setmaterial_main()
{
   GLfloat paintcolor[] = { 1.0, 0.9, 0.8, 1.0 };
   GLfloat specularcolor[] = { 1.0, 1.0, 1.0, 1.0 };
   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, paintcolor);
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularcolor);
   glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.);
}


// drawshadowreceiver
// Draws shadow receiver (at origin)
void drawshadowreceiver()
{
   glutSolidIcosahedron();
}


// drawshadowmaker_placed
// Draws shadow maker, placed within scene
// (not at origin)
void drawshadowmaker_placed()
{
   glPushMatrix();
      glTranslated(0., 0., -4.);
      glutSolidTeapot(1.);
   glPopMatrix();
}


// setupshadowtexture
// Set up shadow texture.
// Must be called when drawing can be done
//    (in display function?).
void setupshadowtexture()
{
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   gluPerspective(60, 1., 0.1, 20.);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   gluLookAt(light0_position[0], light0_position[1], light0_position[2],
             shreceiverpos[0], shreceiverpos[1], shreceiverpos[2],
             0., 1., 0.);

   glPushAttrib(GL_COLOR_BUFFER_BIT);  // Save clear color (among other things)
   glClearColor(1., 1., 1., 1.);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glPopAttrib();                      // Restore clear color

   glColor3d(0., 0., 0.);
   glDisable(GL_LIGHTING);
   drawshadowmaker_placed();
   glEnable(GL_LIGHTING);

   glReadPixels(0, 0, img_width, img_height, GL_RGB, GL_UNSIGNED_BYTE, the_image);

   // Set up texture parameters for texture 0
   glBindTexture(GL_TEXTURE_2D, texnames[0]);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img_width, img_height,
                0, GL_RGB, GL_UNSIGNED_BYTE, the_image);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
}


// display
// The GLUT display function
void display()
{
   // Generate shadow texture, if necessary
   static bool first = true;
   if (first)
   {
      setupshadowtexture();
      first = false;
   }

   // Clear buffers
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // Lighting
   glLightfv(GL_LIGHT0, GL_POSITION, light0_position);

   // Bind Texture
   glBindTexture(GL_TEXTURE_2D, texnames[whichtexture]);
   if (textureon)
      glEnable(GL_TEXTURE_2D);
   else
      glDisable(GL_TEXTURE_2D);

   // Draw shadow receiver

   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();
   glScaled(0.5, 0.5, 0.);  // Turn s,t from [-1,1] to [0,1]
   glTranslated(1., 1., 0.);
   gluPerspective(60, 1., 0.1, 20.);  // Same transformations as used
                                      //  to draw texture image
   gluLookAt(light0_position[0], light0_position[1], light0_position[2],
             shreceiverpos[0], shreceiverpos[1], shreceiverpos[2],
             0., 1., 0.);
   glMatrixMode(GL_MODELVIEW);

   glEnable(GL_TEXTURE_GEN_S);
   glEnable(GL_TEXTURE_GEN_T);
   glEnable(GL_TEXTURE_GEN_R);

   setmaterial_main();

   glPushMatrix();
      glTranslated(shreceiverpos[0],
                   shreceiverpos[1],
                   shreceiverpos[2]);
      glRotated(angle, 0,1,0);
      glScaled(3., 3., 3.);
      drawshadowreceiver();
   glPopMatrix();

   glDisable(GL_TEXTURE_GEN_S);
   glDisable(GL_TEXTURE_GEN_T);
   glDisable(GL_TEXTURE_GEN_R);

   // Draw textured square

   if (drawsquare)
   {
      glDisable(GL_LIGHTING);
      glPushMatrix();
         glTranslated(0., 0., -2.);
         glMatrixMode(GL_TEXTURE);
         glPushMatrix();
            glLoadIdentity();
            glColor3d(1., 1., 1.);
            glBegin(GL_QUADS);
               glTexCoord2d(0., 0.); glVertex3d(-1., -1., 0.);
               glTexCoord2d(1., 0.); glVertex3d( 1., -1., 0.);
               glTexCoord2d(1., 1.); glVertex3d( 1.,  1., 0.);
               glTexCoord2d(0., 1.); glVertex3d(-1.,  1., 0.);
            glEnd();
         glPopMatrix();
         glMatrixMode(GL_MODELVIEW);
      glPopMatrix();
   }

   glDisable(GL_TEXTURE_2D);  // Don't want to affect colors of remaining stuff

   // Draw shadow maker
   glColor3d(0.8, 0.6, 0.4);
   drawshadowmaker_placed();

   // Draw instructions
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_LIGHTING);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
      glLoadIdentity();
      gluOrtho2D(-1., 1., -1., 1.);
      glColor3d(1.0, 0.0, 1.0);
      printbitmap("T      Toggle texture", -0.9, 0.9);
      printbitmap("R      Toggle rotation", -0.9, 0.8);
      printbitmap("S      Toggle textured square", -0.9, 0.7);
      printbitmap("Y      Cycle texture", -0.9, 0.6);
      printbitmap("Esc    Quit", -0.9, 0.5);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glEnable(GL_LIGHTING);
   glEnable(GL_DEPTH_TEST);

   glutSwapBuffers();
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
    glViewport(0, 0, w, h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(90, double(w)/h, neardist, fardist);

    glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case 't':
   case 'T':     // T: toggle texture
      textureon = !textureon;
      glutPostRedisplay();
      break;
   case 'y':
   case 'Y':     // Y: cycle texture
      whichtexture = (whichtexture+1)%numtextures;
      glutPostRedisplay();
      break;
   case 'r':
   case 'R':     // R: toggle rotation
      rotate = !rotate;
      glutPostRedisplay();
      break;
   case 's':
   case 'S':     // S: toggle textured square
      drawsquare = !drawsquare;
      glutPostRedisplay();
      break;
   }
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }

   // Do rotation
   if (rotate)
   {
      angle += anglestep;
      glutPostRedisplay();
   }
}


// makeimage_old
// Make image in the_image
// (Code taken from drawimage.cpp)
void makeimage_old()
{
   for (int i=0; i<img_width; ++i)
   {
      double x = double(i)/(img_width-1);
      for (int j=0; j<img_height; ++j)
      {
         double y = double(j)/(img_height-1);

         the_image[j][i][0] = int(x*255)*15 % 256;
         the_image[j][i][1] = int(y*255)*15 % 256;
         the_image[j][i][2] = 0.75*255;
      }
   }
}


// init
// Initialization
// Called by main after window creation
void init()
{
   glClearColor(0.2, 0.3, 0.4, 1.);
   glEnable(GL_DEPTH_TEST);

   // ***** Texture set-up

   // Basic image set-up
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   // Generate texture names
   glGenTextures(numtextures, texnames);

   // Set up texture parameters for texture 1
   glBindTexture(GL_TEXTURE_2D, texnames[1]);
   makeimage_old();
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img_width, img_height,
                0, GL_RGB, GL_UNSIGNED_BYTE, the_image);

   // Global texture stuff
   glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

   // Texture coordinate generation
   // We want texture coordinates to be the same as vertex coordinates
   GLfloat sparams[] = { 1., 0., 0., 0. };
   GLfloat tparams[] = { 0., 1., 0., 0. };
   GLfloat rparams[] = { 0., 0., 1., 0. };
   glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR); 
   glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR); 
   glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR); 
   glTexGenfv(GL_S, GL_EYE_PLANE, sparams);
   glTexGenfv(GL_T, GL_EYE_PLANE, tparams);
   glTexGenfv(GL_R, GL_EYE_PLANE, rparams);

   // ***** Lighting
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_maincolor);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light0_maincolor);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambcolor);
      // Light 0 position is set in display
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHTING);

   // We do lighting and scaling, so:
   glEnable(GL_NORMALIZE);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Demo of Shadows via Textures");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutIdleFunc(idle);

   // Do something
   glutMainLoop();

   return 0;
}
